import datetime as _datetime
import openpyxl as _openpyxl
import pandas as _pd


#---------------------
# DF and dates manipulation
#---------------------

def normalise_df(df,base=100):
    normalised = df.copy()
    for col in normalised.columns:
        start_value = normalised[col][0]
        normalised[col]=normalised[col]*base/start_value
    return normalised

def excel_date(date1):
    temp = _datetime.date(1899, 12, 30)    # Note, not 31st Dec but 30th!
    delta = date1 - temp
    return float(delta.days) + (float(delta.seconds) / 86400)

def parse_dates_quickly_strict(s, formato = '%Y-%m-%d'):
    """
    s is a time series in the form of DF['date column']
    If not specified, formato = YYYY-MM-DD format
    """
    dates = {date:_pd.to_datetime(date, format=formato) for date in s.unique()}
    return s.map(dates)

#---------------------
# EXCEL READ AUX FUNCTIONS ¦
#---------------------

def read_named_range_from_xlsx(xlsx_file, range_name, has_labels=True, first_col_index=False):
    """
    :param xlsx_file: Full drive path, using forward slash ("/") if needed
    :param range_name: can be a standard Excel reference ('Sheet1!A2:B7') or named region ('ADR')
    :param has_labels: True takes the first row as the column headers
    :param first_col_index:  True makes the first column as the index
    :return: panda df as per the above
    NOTE1: IT DOESN'T WORK OUT OF THE BOX FOR EXCEL TABLES - WE NEED TO DEFINE A NEW NAME
           OVERLAPPING THE TABLE WITH AN EXTRA ROW, SO THAT ITS DEFINITION IS IN THE FORM
           OF "SHEET1!$A3:B7" OR SIMILAR
    NOTE2: IT DOESN'T WORK WITH PASSWORD-PROTECTED BOOKS
    """

    # EXAMPLE CALL
    # from DB_functions_store_retrieve_data import read_named_range_from_xlsx
    #
    # df = read_named_range_from_xlsx(
    #     xlsx_file='C:/Users/juanf/Dropbox/_PROGRAMMING/BBG_python/SISTEMA/SISTEMA_OUTPUT/Book1JuanNP.xlsm',
    #     range_name='ADRs_named_range',
    #     has_labels=True,
    #     first_col_index=False)

    wb = _openpyxl.load_workbook(xlsx_file, data_only=True, read_only=True)
    if '!' in range_name:
        # passed a worksheet!cell reference
        ws_name, reg = range_name.split('!')
        if ws_name.startswith("'") and ws_name.endswith("'"):
            # optionally strip single quotes around sheet name
            ws_name = ws_name[1:-1]
        region = wb[ws_name][reg]
    else:
        # passed a named range; find the cells in the workbook
        full_range = wb.defined_names[range_name]
        if full_range is None:
            raise ValueError(
                'Range "{}" not found in workbook "{}".'.format(range_name, xlsx_file)
            )
        # convert to list (_openpyxl 2.3 returns a list but 2.4+ returns a generator)
        destinations = list(full_range.destinations)
        if len(destinations) > 1:
            raise ValueError(
                'Range "{}" in workbook "{}" contains more than one region.'
                .format(range_name, xlsx_file)
            )
        ws, reg = destinations[0]
        # convert to worksheet object (_openpyxl 2.3 returns a worksheet object
        # but 2.4+ returns the name of a worksheet)
        if isinstance(ws, str):
            ws = wb[ws]
        region = ws[reg]
    # headers and data
    if has_labels:
        cols = [cell.value for cell in region[0]]
        df = _pd.DataFrame(data=[[cell.value for cell in row] for row in region[1:]], columns=cols)
    else:
        df = _pd.DataFrame(data=[[cell.value for cell in row] for row in region[1:]])
    # index
    if first_col_index:
        df.set_index(df.columns[0], inplace=True)
    # return
    return df


def read_ADR_table(file_path='', range_name='',
                   selected_fields=['Source instrument', 'BBG', 'Price multiplier', 'Crncy', 'Crncy dividend', 'Type',
                                    'Cap gains treatment', 'Duration_Nominal', 'Duration', 'Duration_Yield']):
    """
    :param file_path: if we leave it as '', it will read from the clipboard; else put fullpath in there with forward slashes
    :param range_name: the specific range we're targetting in the spreadsheet
    :param selected_fields: the ones we need are already set by default
    :return: df with adr table, indexed by integers and showing the selected fields
    """

    if (file_path == '') or (range_name == ''):
        adr = _pd.read_clipboard(sep='\t', thousands=',')
    else:
        adr = read_named_range_from_xlsx(xlsx_file=file_path, range_name=range_name, has_labels=True,
                                         first_col_index=False)
    if len(selected_fields) > 0:
        adr = adr[selected_fields]
    adr.dropna(how='all', inplace=True)
    # engine = create_SQLalchemy_connection(dbFile=blotterDatabasePath)
    # adr.to_sql(name='adr_table', con=engine, index=False, if_exists='replace')
    # print('Done...saved in DB and results in adr')
    return adr




from optparse import OptionParser as _OptionParser
import pandas as _pd

#Reviewed aux functions:

def parseCmdLine():
    parser = _OptionParser(description="Retrieve reference data.")
    parser.add_option("-a",
                      "--ip",
                      dest="host",
                      help="server name or IP (default: %default)",
                      metavar="ipAddress",
                      default="localhost")
    parser.add_option("-p",
                      dest="port",
                      type="int",
                      help="server port (default: %default)",
                      metavar="tcpPort",
                      default=8194)

    (options, args) = parser.parse_args()

    return options



def groupOutput(listSecID,listFieldID,listFieldScale,oneTablePer,response,onlyCommonDates=True,fillna=False,fillna_method="ffill"):
    '''
    :param listSecID: as per the BDH definition
    :param listFieldID:  as per the BDH definition
    :param listFieldScale:  as per the BDH definition
    :param oneTablePer:  as per the BDH definition
    :param response:  as per the BDH output (a list of [  [securityname, df for that security] ] )
    :return: list of DFs each organised as requestd, by FIELD or by SECURITY
    '''
    fieldAndScale=list(zip(listFieldID,listFieldScale))
    groupedResponse = []


    if oneTablePer=="FIELD":
        # one table = one field, multiple securities
        for (field,scale) in fieldAndScale:
            # ******************************
            # organise list of relevant dates
            df = response[0][1][['date']].copy().set_index('date')  # date
            for response_item in response:
                if onlyCommonDates == True:
                    df.index = df.index.intersection(response_item[1][['date']].copy().set_index('date').index)
                else:
                    df.index = df.index.union(response_item[1][['date']].copy().set_index('date').index)
            df = df.sort_index(ascending=True)
            #******************************
            for security in response:
                sec_brief = security[1][['date', field]].copy().set_index('date')
                sec_brief = sec_brief.rename(columns={field: security[0]})
                #zzzzzzz if by joining the new frame we would get an empty frame, then don't do it
                if onlyCommonDates:
                    df_test=df.join(sec_brief, how='inner')
                else:
                    df_test = df.join(sec_brief, how='outer')
                if df_test.empty==False:
                    df=df_test
                    if fillna:
                        df = df.fillna(method=fillna_method)
                else:
                    print('empty frame alert')
                #zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz
            separator = ("" if scale == "" else "_")
            df.index.rename(field+separator+scale, True)
            groupedResponse.append(df)
        a = 0
        for (field,scale) in fieldAndScale:
            try:
                if scale == "REBASE100":
                    for securityName in listSecID:
                        baseValue = groupedResponse[a][securityName].dropna()[0]/100
                        groupedResponse[a].loc[:, securityName] /= baseValue
                if scale=="RETURNS":
                    groupedResponse[a]=groupedResponse[a].pct_change(1)
            except:
                print(f"cannot rebase {field}")
            a=a+1
    else:
        # one table = one security, multiple fields
        for security in response:  #this should provide them in the same order
            try:
                df = security[1].set_index('date')
                # ****new stuff****
                a=0
                for (field,scale) in fieldAndScale:
                    separator=("" if scale=="" else "_")
                    df.columns.values[a]=field+separator+scale
                    a=a+1
                #******************
                df.index.rename(security[0], True)
                if fillna:
                    df = df.fillna(method=fillna_method)
                groupedResponse.append(df)
            except:
                df = None
        a=0
        for table in groupedResponse:
            if isinstance(table, _pd.DataFrame):
                for (field,scale) in fieldAndScale:
                    separator=("" if scale=="" else "_")
                    if scale == "REBASE100":
                        groupedResponse[a].loc[:, field+separator+scale] /= (groupedResponse[a][field+separator+scale][0]/100)
                    if scale=="RETURNS":
                        groupedResponse[a][field+separator+scale]=groupedResponse[a][field+separator+scale].pct_change(1)
            a=a+1


    return groupedResponse

#---------------------
# TREE AUX FUNCTIONS ¦
#---------------------

def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]










import _datetime
strptime = _datetime.datetime.strptime

def list_results(result,tab="___"):
    if not (isinstance(result,list) or isinstance(result,dict) or isinstance(result,tuple)):
        print(tab,result)
        return
    if isinstance(result,list) or isinstance(result,tuple):
        for listElement in result:
            list_results(listElement)
        return
    for key in result:
        value = result[key]
        if isinstance(value,dict):
            print(tab, key)
            newtab=tab+"___"
            list_results(value,newtab)
        elif isinstance(value,list):
            print(tab, key)
            newtab = tab + "___"
            for listElement in value:
                list_results(listElement,newtab)
        else:
            print(tab,key,":",result[key])

def date_interval_iso(num_days=0,startDateYYYYMMDD="",endDateYYYYMMDD=""):
    '''
    :param num_days:
    :param startDateYYYYMMDD:
    :param endDateYYYYMMDD:
    :return: if either start or end dates are specified, it will return start, end adjusted by num_days. If neither start or end are specified, then it will be number of days from today, backwards
    '''

    def isoformat(date):
        return date.isoformat().replace("-", "")

    delta = _datetime.timedelta(days=num_days)
    if (startDateYYYYMMDD=="" and endDateYYYYMMDD==""):
        today = _datetime.date.today()
        end = isoformat(today)
        start = isoformat(today-delta)
    elif startDateYYYYMMDD!="":
        end = isoformat(strptime(startDateYYYYMMDD,"%Y%m%d")+delta)[0:8]
        start = startDateYYYYMMDD
    elif (startDateYYYYMMDD!="" and endDateYYYYMMDD!=""):
        start = startDateYYYYMMDD
        end = endDateYYYYMMDD
    else:
        start = isoformat(strptime(endDateYYYYMMDD,"%Y%m%d") - delta)[0:8]
        end = endDateYYYYMMDD

    return [start,end]



import anytree
import pandas as pd
from anytree import RenderTree, findall, ZigZagGroupIter
from ..BBG.core import BDH

pd.set_option('display.max_columns', None)

class Node(anytree.Node):
    '''
    THIS EXPANDS THE NODE CLASS IN THE ANYTREE LIBRARY WITH THE FOLLOWING FUNCTIONALITY:
    adding: node1 + node2 will calculate the union sets and return a tuple with the resulting (asset_ids,asset_names)
    subtracting: node1 - node2 will return a tuple of the (asset_ids, asset_names) in node1 so long as they dont appear in node2
    render: node.render() will perform a basic rendering of such node, all the way downwards
    '''
    def __add__(self,other):
        asset_ids=self.asset_ids.copy()
        for asset_id in other.asset_ids:
            if not(asset_id in asset_ids):
                asset_ids+=[asset_id]
        temp_node = Node(name="temp",parent=None)
        temp_node.asset_ids = asset_ids
        return temp_node

    def __sub__(self, other):
        asset_ids=[]
        for asset_id in self.asset_ids:
            if not(asset_id in other.asset_ids):
                asset_ids+=[asset_id]
        temp_node = Node(name="temp",parent=None)
        temp_node.asset_ids = asset_ids
        return temp_node


    def render(self,sort=True,show_aggregate = False,aggregate_digits=7):

        #define sort function - whether by name of the node or, if we have additional_printout, by additional_printout
        def mysort(items):
            if sort:
                if show_aggregate:
                    return sorted(items, key=lambda item: item.aggregate, reverse=True)
                else:
                    return sorted(items, key=lambda item: item.name, reverse=False)
            else:
                return items

        #iterate
        for pre, _, node in RenderTree(self,childiter=mysort):
            if show_aggregate:
                exec(f"global _addon;_addon = ''+str(node.aggregate)+'  '")
                _addon_format = "%" + str(int(aggregate_digits)) + "s"
            else:
                _addon_format="%s"
                exec(f"global _addon;_addon = ''")
            print(("%s"+_addon_format+"%s") % (pre, _addon,node.name))

    def search(self, node_name, partial_match=False, first_result=True, verbose=False):
        '''
        :param node_name: string to search for
        :param partial_match: if True, will return nodes that contain the node_name. If false, it will only bring exact matches
        :param first_result: if True, only returns one result
        :param verbose: if True, will print number of search results alongside asset count
        :return: single Node or tuple of Nodes depending on the options. If no results found, then None.
        '''
        if partial_match:
            found = findall(self, filter_=lambda node: node_name in node.name)
        else:
            found = findall(self, filter_=lambda node: node.name in [node_name])
        if verbose:
            print(f"{len(found)} results found")
            for node in found:
                print(node.name + f"---- {len(node.asset_ids)} assets")
            if len(found)>1 and first_result:
                print("returning first match only")
        if len(found)>0:
            if first_result or len(found)==1:
                return found[0]
            else:
                return found
        else:
            return None

    def custom_aggregate(self,aggregate_field="",aggregate_function="sum", filter_fields=[], filter_values=[],propagate=False,verbose=True):
        '''
        :param aggregate_field: eg "max_occupancy", must be one of numeric fields in the configuration_df (--> INTERNAL CONFIG DATA!!!)
        :param aggregate_function: "sum", "max", etc.
        :param filter_fields: [list of configuration df columns to filter by]
        :param filter_values:[list of filter values]
        :param propagate: if True, it will create for this node and all branches below a new property node.aggregate with the outcome of the computation at node level (eg max capacity)
        :return: aggregate for the parent node (although it will also have updated node.aggregate and all the subnodes.aggregate. It will also inherit data through all subnodes!
        '''

        #if propagate is True we create a tuple with levels of nodes
        if propagate==True:
            zigzag = [node for node in ZigZagGroupIter(self)]
        else:
            zigzag = [[self]]
        for level in reversed(zigzag):
            for node in level:
                # ensure it has data (NOTE - THIS WILL ONLY WORK IF DATA HAS BEEN LOADED FIRST)
                if node != self:
                    try:
                        node.inherit_data(self)
                    except:
                        pass
                #calculate custom aggregate
                df = node.root.configuration_df.copy()
                df = df[(df[node.asset_id_field].isin(node.asset_ids))]
                for filter_name,filter_value in zip(filter_fields,filter_values):
                    df = df[df[filter_name]==filter_value]
                exec ("global _custom_aggregate_output;_custom_aggregate_output = df['"+aggregate_field+"']."+aggregate_function+"()")
                if verbose:
                    print(f"{node.name} {aggregate_field} {aggregate_function} = {_custom_aggregate_output}")
                if propagate:
                    node.aggregate = _custom_aggregate_output

        return _custom_aggregate_output


    def load_params(self,params_dict):
        '''
        :param params_dict: format as per the below
            params_dict = {
                "listSecID":["AAPL US Equity"],
                "listFieldID":["PX_LAST","BEST_PE_RATIO"],
                "listFieldScale":[],
                "startDateYYYYMMDD":"20220701",
                "endDateYYYYMMDD":"20230131",
                "periodicityAdjustment":"ACTUAL","periodicitySelection":"DAILY","maxDataPoints":0,
                "overrides":[],
                "convertDatesToNumbers":False,
                "oneTablePer":"SECURITY",
                "onlyCommonDates":True,
                "return_as_dict":False
                }
        :return: it creates a self.params attribute that can be accessed further, for instance self.params.startDateYYYYMMDD.
        It also creates for each telemetry a self.data.TEMPERATURE (for instance) where the list of downloaded dfs per asset will be stored
        '''
        self.params = Params(params_dict)
        self.data = Data(self.params,len(self.asset_ids))
        return self


    def load_data(self):
        '''
        needs to have loaded params first!
        assumes one table per security & concat = False
        :return: each telemetry will have a list of dfs corresponding to the data downloaded for each asset in self.asset_ids
        '''
        temp_df = BDH(
            listSecID=self.params.listSecID,
            listFieldID=self.params.listFieldID,
            listFieldScale=self.params.listFieldScale,
            startDateYYYYMMDD=self.params.startDateYYYYMMDD,
            endDateYYYYMMDD=self.params.endDateYYYYMMDD,
            periodicityAdjustment=self.params.periodicityAdjustment,
            periodicitySelection=self.params.periodicitySelection,
            listOverrides=[],
            maxDataPoints = self.params.maxDataPoints,
            convertDatesToNumbers = self.params.convertDatesToNumbers,
            oneTablePer="SECURITY",
            onlyCommonDates=self.params.onlyCommonDates,
            return_as_dict=False)

        for df in temp_df:
            for field in self.params.listFieldID:
                relevant_cols = [col for col in temp_df.columns if ("_"+field) in col]
                temp_df_metric = temp_df.copy()[relevant_cols]
                temp_df_metric.columns = ["_".join(x.split("_")[:-1]) for x in temp_df_metric.columns]
                exec("self.data." + field + "[" + str(i) + "]=temp_df_metric")
                if len(temp_df_metric)==0:
                    print(f"Data not found for asset {asset_id} ({field}")
            return self

    def inherit_data(self,data_node):
        self.params = data_node.params
        self.data = Data(self.params,len(self.asset_ids))
        for telemetry in [attribute for attribute in dir(data_node.data) if not attribute.startswith('__')]:
            exec(f"parent_dict = data_node.{telemetry}")
            for i,asset_id in zip(range(0,len(self.asset_ids)),self.asset_ids):
                if asset_id in data_node.asset_ids:
                    exec("self.data."+telemetry+"[i] = parent_dict[asset_id]")


    def __if_no_children_then_condense(space, telemetry, operation):
        if len(space.children) == 0:
            sum_of_assets = space.merge_data(telemetry, operation)
            sum_of_assets.columns = [space.name + "_condensed"]

            space.condensed = [sum_of_assets]
        else:
            space.condensed = []
            for child in space.children:
                space.condensed += child.condensed

    def condense(root_node, telemetry, operation):
        '''
        :param root_node: the top level node we want to condense all the way down
        :param telemetry: "AREA_COUNT" etc
        :param operation: "sum", "max", "min", "etc"
        :return: it modifies every node from root_node downwards by adding the .condensed property with a df ofor each of the leaf spaces (not assets) under it
        '''
        #this creates a tuple with levels of nodes
        zigzag = [node for node in ZigZagGroupIter(root_node)]
        for level in reversed(zigzag):
            for node in level:
                # ensure it has data
                if node != root_node:
                    node.inherit_data(root_node)

                #if there are only assets under it, not further children spaces, then it aggregates all asset data into one
                if len(node.children) == 0:
                    sum_of_assets = node.merge_data(telemetry, operation)
                    sum_of_assets.columns = [node.name + "_condensed"]
                    node.condensed = [sum_of_assets]
                else:
                    #if it has space children, it creates a list of all the condensed spaces below it
                    node.condensed = []
                    for child in node.children:
                        node.condensed += child.condensed

    def merge_data(self, telemetry, transform_function="plain", use_asset_names=False, filter_fields=[], filter_values=[]):
        '''
        :param telemetry: single telemetry (eg "AREA_COUNT")
        :param transform_function: "plain" if we want asset by asset results, any of the pandas functions ("sum", "min", "max", "median" etc) if not. Applied horizontally to all assets.
        :param use_asset_names: if True, it will return asset names instead of asset ids as columns
        :param configuration_df: optional - if we want to filter by attribute name and value, this df will contain asset_id plus the attribute names
        :param filter_fields: list of attribute names (columns of configuration_df) we want to filter by
        :param filter_values: list of attribute values we want to filter by. All need to match concurrently.
        :return: df with the resulting data, and where the columns are either asset names ("plain") or telemetry_operation (eg "AREA_COUNT_sum")
        '''
        global _combined_df
        df_list=[]
        for asset_id, asset_df in zip(self.asset_ids,self.__text_to_property(telemetry)):
            if asset_df is None:
                pass
            else:
                df_list+=[asset_df]

        #*****30-08-22 add filtering*****
        configuration_df = self.root.configuration_df
        if configuration_df is None:
            pass
        else:
            for attribute_name, attribute_value in zip(filter_fields, filter_values):
                configuration_df = configuration_df[configuration_df[attribute_name]==attribute_value]
            eligible_assets = [str(asset) for asset in list(configuration_df["asset_id"])]
            df_list = [df for df in df_list if str(df.columns[0]) in eligible_assets]
        # *****30-08-22 add filtering*****

        if len(df_list)>0:
            _combined_df = pd.concat(df_list,axis=1)
            _combined_df.index.name=telemetry
            if transform_function!="plain":
                exec("global _combined_df;_combined_df = _combined_df."+transform_function+"(axis=1).to_frame()")
                _combined_df.columns=[telemetry+"_"+transform_function]
            if use_asset_names and transform_function=="plain":
                cols = list(_combined_df.columns)
                new_cols = []
                id_to_name_dictionary = res = {str(self.asset_ids[i]): self.asset_names[i] for i in range(len(self.asset_ids))}
                for i in range(0,len(cols)):
                    new_cols+=[id_to_name_dictionary[str(cols[i])]]
                _combined_df.columns=new_cols
            return _combined_df
        else:
            return None

    def __text_to_property(self,property_name):
        print(self.name)
        exec("global temp;temp=self.data."+property_name)
        return temp


    # *******************11-09-2022**************

class Params:
    def __init__(self,params_dict):
        self.params_dict = params_dict
        for parameter in params_dict:
            exec("self."+parameter+"=params_dict[parameter]")
class Data:
    def __init__(self,params,num_assets):
        for field in params.listFieldID:
            exec("self."+field+"=[None]*"+str(num_assets))



def create_tree(configuration_df,levels,initial_depth,asset_id_field="",add_asset_ids=False):
    '''
    :param configuration_df: a df containing asset names and ids plus whatever fields we deem appropriate for accordion classification. Must have the columns asset_name and asset_id
    :param levels: a list with the names of the grouping columns we want to generate the tree from, eg levels=["building","group","level_01","level_02","level_03"]
    :param initial_depth: the level we want to start the tree at (e.g. in the example above depth=0 starts with "building"
    :param add_asset_ids: if True, it will add asset ID at the end of each node
    :return: a list of the root Nodes of the tree
    '''
    global id
    id = 0
    global return_nodes
    return_nodes=[]
    def __recursive_tree(df,levels,initial_depth,parent=None,add_asset_ids=False):
        global id
        depth=initial_depth
        level=levels[0:depth+1]
        idx = df.groupby(level,dropna=True).indices
        if (len(idx)==0) and (add_asset_ids==True):
            for asset_id in parent.asset_ids:
                id=id+1
                asset_node = Node(name=str(asset_id),parent=parent)
                asset_node.asset_id_field = asset_id_field
                exec("global ID"+str(id)+";ID"+str(id)+"=asset_node")

        for item in idx:
            if type(item)==str:
                new_node=item
            else:
                new_node=item[-1]
            new_df=df[df[levels[depth]]==new_node]
            if asset_id_field!="":
                asset_id_list = list(new_df[asset_id_field])
            else:
                asset_id_list=[]
            node = Node(new_node,
                        parent=parent,
                        asset_ids=asset_id_list)
            node.asset_id_field = asset_id_field
            exec("global ID"+str(id)+";ID"+str(id)+"=node")
            if parent==None:
                exec("global ID"+str(id)+";return_nodes+=[ID"+str(id)+"]")

            id=id+1
            if (depth+1)<len(levels):
                __recursive_tree(df=new_df,levels=levels,initial_depth=depth+1,parent=node,add_asset_ids=add_asset_ids)

    __recursive_tree(df=configuration_df,levels=levels,initial_depth=initial_depth,parent=None,add_asset_ids=add_asset_ids)
    # 2022-09-10
    for node in return_nodes:
        node.configuration_df = configuration_df
        node.levels = levels
    #***********
    return return_nodes






from ..BBG.core import BDH
from .general import date_interval_iso as _date_interval_iso
import plotly.graph_objects as _go
import plotly.express as _px
from plotly.subplots import make_subplots as _make_subplots
import _datetime
strptime = _datetime.datetime.strptime
import pandas as pd

def plotly_subplots(rows=1, cols=1):
    '''
    :param rows: number of rows in the plot
    :param cols: number of columns in the plot
    :return: figure with row,column suplots AND secondary axis enabled

    example: fig = plotly_subplots(2,1)
    It can then be used in combination with chart_by_axis or download_and_chart
    '''

    specs = []
    for i in range(0, rows):
        specs += [[{"secondary_y": True}] * cols]
    fig = _make_subplots(specs=specs, rows=rows, cols=cols)
    return fig


def plotly_download_and_chart(listSecID,listFieldID,yaxis=[[]],num_days=10,startDateYYYYMMDD="",endDateYYYYMMDD="",fillna=True,target_fig=None,row=1,col=1):
    '''
    :param listSecID: list of securities
    :param listFieldID: list of fields
    :param yaxis: double brackets: specify 1 or 2 depending on where we want each chart. Order =
            [
                [sec1field1, sec2field1, sec3field1]
                [sec1field2, sec2field2, sec3field2]
                etc
            ]
    :param num_days: we can specify start and end, or start + num days, or end + num days
    :param startDateYYYYMMDD:
    :param endDateYYYYMMDD:
    :return: returns plotly fig and data as a dictionary of dfs by field
    '''
    start, end = _date_interval_iso(num_days,startDateYYYYMMDD,endDateYYYYMMDD)
    data = BDH(listSecID=listSecID,listFieldID=listFieldID,listFieldScale=[],startDateYYYYMMDD=start,endDateYYYYMMDD=end,oneTablePer="FIELD", return_as_dict=True,onlyCommonDates=False,fillna=fillna)
    print(startDateYYYYMMDD, endDateYYYYMMDD)

    if target_fig==None:
        fig=plotly_subplots(rows=1,cols=1)
    else:
        fig = target_fig

    colors = _px.colors.qualitative.Plotly[0:len(listSecID)*len(listFieldID)]
    colors = [colors[i:i + len(listSecID)] for i in range(0, len(colors), len(listSecID))]
    colors = {field:color for field,color in zip(listFieldID,colors)}
    if yaxis==[[]]:
        yaxis= {field: [len(listSecID)*False] for field in listFieldID}
    else:
        temp = {}
        for field,row in zip(listFieldID,yaxis):
            temp[field]= [True if yaxe == 2 else False for yaxe in row]
        yaxis=temp

    for field in listFieldID:
        df = data[field]
        field_axis = yaxis[field]
        field_colors = colors[field]
        for sec,axis,color in zip(listSecID,field_axis,field_colors):
            fig.add_trace(
                _go.Scatter(
                    x=df.index,
                    y=df[sec],
                    name=sec+"_"+field,
                    line_color=color),
                secondary_y= axis,
                row=row,
                col=col
            )
    return fig,data

def plotly_chart_by_axis(df_or_list_of_series=[],yaxis=[],num_days=10,startDateYYYYMMDD="",endDateYYYYMMDD="",fillna=True,target_fig=None,row=1,col=1):
    '''
    :param df_or_list_of_series: either a df or a list with each element being a pandas series
    :param yaxis: specify 1 or 2 depending on where we want each chart. Order =
                [series1, series2, series3...]
    :param num_days: we can specify start and end, or start + num days, or end + num days
    :param startDateYYYYMMDD:
    :param endDateYYYYMMDD:
    :return: returns plotly fig
    '''
    if isinstance(df_or_list_of_series,pd.DataFrame):
        list_of_series = [df_or_list_of_series[series] for series in df_or_list_of_series]
    else:
        list_of_series=df_or_list_of_series

    if target_fig==None:
        fig=plotly_subplots(row=row,col=col)
    else:
        fig = target_fig

    colors = _px.colors.qualitative.Plotly[0:len(list_of_series)]
    if yaxis==[]:
        yaxis= [False]*len(list_of_series)
    else:
        yaxis = [True if yaxe == 2 else False for yaxe in yaxis]

    for series,axis,color in zip(list_of_series,yaxis,colors):
        fig.add_trace(
            _go.Scatter(
                x=series.index,
                y=series,
                name=series.name,
                line_color=color),
            secondary_y= axis,
            row=row,
            col=col
        )
    return fig


import setuptools

# Reads the content of your README.md into a variable to be used in the setup below
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name='JFQ',  # should match the package folder
    packages=setuptools.find_packages(),  # should match the package folder
    version='v0.1.9',  # important for updates
    license='GNU2L',  # should match your chosen license
    description='various utilities',
    long_description=long_description,  # loads your README.md
    long_description_content_type="text/markdown",  # README.md is of type 'markdown'
    author='',
    author_email='',
    url='', #removed github url
    install_requires=['requests','pandas','matplotlib','xlsxwriter','anytree','blpapi','openpyxl'],  # list all packages that your package uses
    keywords=["openAPI"],  # descriptive meta-data
    classifiers=[  # https://pypi.org/classifiers
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Documentation',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11'
    ]
)


from __future__ import print_function
from __future__ import absolute_import
import blpapi as _blpapi
import pandas as _pd
import datetime as _datetime
from ..HELPERS import misc as _misc
from ..HELPERS import pandas_ext as _pandas_ext
import numpy as _np

#FULLY REVIEWED:
#BDH
#BDH_Disjoint
#BEQS
#BDP_One_Field --> improved to show field name in the index, and the ability to display securities as columns or as rows
#BDP_Multiple_Fields
#BDP_Dividends
#BDP_Bond_Cashflows -> improved to be able to specify face amount. Consider in the future simplifying the override structure to a settlement date parameter
#generate_universe -> improved and merged so that it returns both the universe as well as the unit and raw weights


def BDH(listSecID,listFieldID,listFieldScale,startDateYYYYMMDD,endDateYYYYMMDD,periodicityAdjustment="ACTUAL",periodicitySelection="DAILY",maxDataPoints=0,listOverrides = [],convertDatesToNumbers=False,oneTablePer="SECURITY",onlyCommonDates=True,return_as_dict=False,fillna=False,fillna_method="ffill"):
    '''
    Calls the historic data download function from bloomberg
    :param listSecID: list of the tickers we want to use, eg ["IBM US Equity", "AAPL US Equity"]
    :param listFieldID: list of the historical fields requested, eg ["PX_LAST"]
    :param listFieldScale: list matching the listFieldID length where each of the items is "RAW", "REBASE100", or "RETURNS". If left as an empty list, it will assume RAW.
    :param startDateYYYYMMDD: text string in YYYYMMDD format, eg "20061230"
    :param endDateYYYYMMDD: text string in YYYYMMDD format, eg "20061230"
    :param periodicityAdjustment: "ACTUAL","CALENDAR", or "FISCAL"
    :param periodicitySelection: "DAILY", "WEEKLY", "MONTHLY","QUARTERLY","SEMI_ANNUALLY","YEARLY"
    :param maxDataPoints: if 0, then no limitation
    :param overrides = list of overrides, where each override itself is a 2-uple of (override field,override value)
    :param convertDatesToNumbers: the default is False; if True, it converts it into a excel number equivalent
    :param oneTablePer: "SECURITY" (default: each table is one security, shows all fields) or "FIELD" (each table is one field, shows result for all securities)
    :param onlyCommonDates: If True, it will restrict df output to dates found in all
    :param return_as_dict: If true, the return will be a dictionary of the form field: table by field or security: table by security
    :param fillna: if True, fillna with fillna_method
    :param fillna_method: "ffill", etc
    :return: returns a list of data_frames (or dictionary, if return_as_dict is True) - if "SECURITY" has been chosen, then it'll be all fields for a given security; otherwise it'll be all securities for a given field.
    '''
    options = _misc.parseCmdLine()

    #if listFieldScale is an empty list
    if listFieldScale==[]:
        listFieldScale=[""]*len(listFieldID)

    # Fill SessionOptions
    sessionOptions = _blpapi.SessionOptions()
    sessionOptions.setServerHost(options.host)
    sessionOptions.setServerPort(options.port)

    print("Connecting to %s:%s" % (options.host, options.port))
    # Create a Session
    session = _blpapi.Session(sessionOptions)

    # Start a Session
    if not session.start():
        print("Failed to start session.")
        return

    try:
        # Open service to get historical data from
        if not session.openService("//blp/refdata"):
            print("Failed to open //blp/refdata")
            return

        # Obtain previously opened service
        refDataService = session.getService("//blp/refdata")

        # Create the request for the historical data
        request = refDataService.createRequest("HistoricalDataRequest")

        # Fill the request
        #   a) securities
        for sec_id in listSecID:
            request.getElement("securities").appendValue(sec_id)
        #   b) fields
        for field_id in listFieldID:
            request.getElement("fields").appendValue(field_id)
        #   c) periodicity
        request.set("periodicityAdjustment", periodicityAdjustment)
        request.set("periodicitySelection", periodicitySelection)
        #   d) start and end date
        request.set("startDate", startDateYYYYMMDD)
        request.set("endDate", endDateYYYYMMDD)
        #   e) max data points
        request.set("maxDataPoints", maxDataPoints)
        #   f) other overrides
        if len(listOverrides)>0:
            overrides = request.getElement("overrides")
            for pair in listOverrides:
                override = overrides.appendElement()
                override.setElement('fieldId', pair[0])
                override.setElement('value', pair[1])

        # Send the request
        if False:
            print("Sending Request:", request)
        else:
            print("Historical Request - "+str(listFieldID)+" - "+startDateYYYYMMDD+" : "+endDateYYYYMMDD)
        session.sendRequest(request)

        # Process received events
        RESPONSE= _blpapi.Event.RESPONSE
        PARTIAL_RESPONSE= _blpapi.Event.PARTIAL_RESPONSE
        allSecurities = []  #list of "securities". Each security is in the form of [sec_name, panda_dataframe with dates and fields]
        while(True):
            # We provide timeout to give the chance for Ctrl+C handling:
            ev = session.nextEvent(10)         #original value=500
            for msg in ev:
                #print(msg)
                if (ev.eventType()==PARTIAL_RESPONSE) or (ev.eventType()==RESPONSE):
                    sd = msg.getElement("securityData")
                    sn = sd.getElement("security").getValue()
                    fd = sd.getElement("fieldData")
                    numDataRows=fd.numValues()
                    try:
                        labels=['date']+listFieldID
                        securityData=[]    #This variable will contain all data rows
                        for a in range(0,numDataRows):#each a "row of data", bbg calls them "fields"
                            dataRow=fd.getValue(a)
                            oneRow = [_np.nan]*len(labels)
                            numFields = dataRow.numElements()
                            for b in range(0,numFields):
                                v=dataRow.getElement(b).getValue(0)
                                if type(v) is _datetime.date and convertDatesToNumbers:
                                    v= _pandas_ext.excel_date(v)
                                # oneRow.append(v)
                                n = str(dataRow.getElement(b).name())
                                if n in labels:
                                    oneRow[labels.index(n)]=v
                            securityData.append(oneRow)
                            #print(row_list)
                        df= _pd.DataFrame.from_records(securityData, columns=labels)
                        # df.set_index('date')
                        #2023-02 force date to be a date
                        if not(convertDatesToNumbers):
                            df["date"]= _pd.to_datetime(df["date"])

                        #---------------------------------
                        if onlyCommonDates:
                            df=df.dropna()

                        oneSecurity=[sn,df]
                        allSecurities.append(oneSecurity)
                    except:
                        print('Empty values for '+sn+' on '+startDateYYYYMMDD+' - '+endDateYYYYMMDD)

            if ev.eventType() == _blpapi.Event.RESPONSE:
                # Response completly received, so we could exit
                break
    finally:
        # Stop the session
        session.stop()
        # at this point allSecurities = [[sec_name, panda_dataframe]], where panda_dataframe is of the format 'date','field 1'...'field n', for a given security

        #sort by listSecID
        sec_df_dict = {sec[0]: sec[1] for sec in allSecurities}
        allSecurities = [ [security,sec_df_dict[security]] if security in sec_df_dict else [security,None] for security in listSecID ]

        # group output and return values
        allSecuritiesGrouped= _misc.groupOutput(listSecID,listFieldID,listFieldScale,oneTablePer,allSecurities,onlyCommonDates,fillna,fillna_method)

        #if return as dict:
        if return_as_dict:
            allSecuritiesGrouped = {table.index.name:table for table in allSecuritiesGrouped}

        return allSecuritiesGrouped

#----------------------------------------------------------------------------------------------------------
#----------------------------------------------------------------------------------------------------------
#----------------------------------------------------------------------------------------------------------


def BDH_Disjoint(listSecID,listFieldID,startDateYYYYMMDD,endDateYYYYMMDD,periodicityAdjustment="ACTUAL",periodicitySelection="DAILY",listOverrides = [],maxDataPoints=0,return_as_dict=False):
    allSecurities = BDH(listSecID=listSecID,
                        listFieldID=listFieldID,
                        listFieldScale=[],
                        startDateYYYYMMDD=startDateYYYYMMDD,
                        endDateYYYYMMDD=endDateYYYYMMDD,
                        periodicityAdjustment=periodicityAdjustment,
                        periodicitySelection=periodicitySelection,
                        maxDataPoints=maxDataPoints,
                        listOverrides=listOverrides,
                        oneTablePer="SECURITY",
                        onlyCommonDates=False,
                        return_as_dict=return_as_dict
                        )
    return allSecurities

def BDH_stacked(listSecID,listFieldID,startDateYYYYMMDD,endDateYYYYMMDD,periodicityAdjustment="ACTUAL",periodicitySelection="DAILY",listOverrides = [],maxDataPoints=0,stacked=True,index_reset=True):
    '''
    :param listSecID: as usual
    :param listFieldID: as usual
    :param startDateYYYYMMDD: as usual
    :param endDateYYYYMMDD: as usual
    :param periodicityAdjustment: as usual, can be ommited in most cases
    :param periodicitySelection: as usual, can be ommited in most cases
    :param listOverrides: as usual, can be ommited in most cases
    :param maxDataPoints: as usual, can be ommited in most cases
    :param stacked: If True (default) --> produces plotly-friendly data sets
    :param index_reset: If True (default) --> produces plotly-friendly data sets, otherwise the index will be (security,date)
    :return: plotly-friendly dataframe
    '''
    response = BDH_Disjoint(listSecID=listSecID, listFieldID=listFieldID, startDateYYYYMMDD=startDateYYYYMMDD, endDateYYYYMMDD=endDateYYYYMMDD, periodicityAdjustment=periodicityAdjustment, periodicitySelection=periodicitySelection, listOverrides=listOverrides,maxDataPoints=maxDataPoints,return_as_dict=False)
    allSecurities=_pd.DataFrame(columns=['security','date',listFieldID[0]],data=None)
    allSecurities.set_index(['security', 'date'], inplace=True)

    for sec in response:
        sec_name=sec.index.name
        sec.index.name='date'
        sec['security']=sec_name
        sec = sec.reset_index().set_index(['security', 'date'])
        allSecurities = _pd.concat([allSecurities,sec],axis=0)

    if stacked==False:
        allSecurities=allSecurities.unstack('security')

    if index_reset:
        allSecurities=allSecurities.reset_index()
    #return array
    return allSecurities

def BEQS(screenName,screenType,asOfYYYYMMDD):
    '''
    :param screenName: name of the pre-saved bloomberg screen
    :param screenType: normally this will be "PRIVATE"
    :param asOfYYYYMMDD: date as of which we want the screen
    :return: returns a list of tickers that meet the screen filters as of that date; no other fields returned
    '''

    session= _blpapi.Session()
    session.start()
    session.openService("//blp/refdata")

    refDataService=session.getService("//blp/refdata")
    request=refDataService.createRequest("BeqsRequest")
    request.set("screenName", screenName)
    request.set("screenType", screenType)
    overrides=request.getElement("overrides")
    override=overrides.appendElement()
    override.setElement('fieldId','PiTDate')
    override.setElement('value' ,asOfYYYYMMDD)
    override = overrides.appendElement()

    session.sendRequest(request)

    endReached = False
    while endReached == False:
        ev = session.nextEvent()
        if ev.eventType() == _blpapi.Event.RESPONSE or ev.eventType() == _blpapi.Event.PARTIAL_RESPONSE:
            for msg in ev:
                dt=msg.getElement("data").getElement("securityData")
                numSecurities=dt.numValues()
                tickers=[]
                for i in range(0,numSecurities):
                    sd=dt.getValue(i)
                    sn = sd.getElement("security").getValue()
                    fd = sd.getElement("fieldData") #we won't do anything with this since we only need the security name
                    tickers.append(sn+" Equity")
        if ev.eventType() == _blpapi.Event.RESPONSE:
            endReached = True
    print('EQS done for '+asOfYYYYMMDD+'!')
    return tickers

    # This is the info returned by default on an EQS - we ignore all but the ticker, we will request the info we need separately via BDH
    #Ticker = "OVAS US"
    #Short Name = "OVASCIENCE INC"
    #Market
    #Cap = 47430628.000000
    #Price: D - 1 = 1.330000
    #Total Return YTD = -13.071891
    #Revenue T12M = 570000.000000
    #EPS T12M = -2.220000

#----------------------------------------------------------------------------------------------------------
#----------------------------------------------------------------------------------------------------------
#----------------------------------------------------------------------------------------------------------



#----------------------------------------------------------------------------------------------------------
#----------------------------------------------------------------------------------------------------------
#----------------------------------------------------------------------------------------------------------
def BDP_One_Field(listSecID,targetFieldName,overrideList,as_columns=True):
    '''
    Calls for one (1) field ref data request (including Total Return), can use multiple securities
    :param listSecID: list of the tickers we want to use, eg ["IBM US Equity", "AAPL US Equity"]
    :param targetFieldName: the field we're looking for (eg "CUST_TRR_RETURN_HOLDING_PER")
    :param overrideList: [[overrideParam,overrideValue]] eg [['CUST_TRR_START_DT',startDateYYYYMMDD],['CUST_TRR_END_DT',endDateYYYYMMDD]]
    :param as_columns: if True, securities will be in the columns field. Otherwise in the rows field.
    :return: returns a dataframe with {securities as column or row headers} and {one row or column with the data for ONE field as requested}
    '''
    options = _misc.parseCmdLine()

    # Fill SessionOptions
    sessionOptions = _blpapi.SessionOptions()
    sessionOptions.setServerHost(options.host)
    sessionOptions.setServerPort(options.port)

    print("Connecting to %s:%s" % (options.host, options.port))
    # Create a Session
    session = _blpapi.Session(sessionOptions)

    # Start a Session
    if not session.start():
        print("Failed to start session.")
        return

    try:
        # Open service to get historical data from
        if not session.openService("//blp/refdata"):
            print("Failed to open //blp/refdata")
            return

        # Obtain previously opened service
        refDataService = session.getService("//blp/refdata")

        # Create the request for the historical data
        request = refDataService.createRequest("ReferenceDataRequest")

        # Fill the request
        #   a) securities
        for sec_id in listSecID:
            request.getElement("securities").appendValue(sec_id)
        #   b) fields
        request.getElement("fields").appendValue(targetFieldName)
        #   d) overrides
        overrides = request.getElement("overrides")
        for ovr in overrideList:
            override = overrides.appendElement()
            override.setElement('fieldId', ovr[0])
            override.setElement('value', ovr[1])

        # Send the request
        if False:
            print("Sending Request:", request)
        else:
            if len(overrideList)>0:
                print("Processing BDP - "+targetFieldName+' - '+str(overrideList[0]))
        session.sendRequest(request)

        # Process received events
        RESPONSE= _blpapi.Event.RESPONSE
        PARTIAL_RESPONSE= _blpapi.Event.PARTIAL_RESPONSE
        tickers=[]
        fieldValues=[]

        while(True):
            # We provide timeout to give the chance for Ctrl+C handling:
            ev = session.nextEvent(10)         #original value=500
            for msg in ev:
                #print(msg)
                if (ev.eventType()==PARTIAL_RESPONSE) or (ev.eventType()==RESPONSE):
                    dt = msg.getElement("securityData")
                    numSecurities = dt.numValues()
                    for i in range(0, numSecurities):
                        sd = dt.getValue(i)
                        sn = sd.getElement("security").getValue()
                        tickers.append(sn)
                        fd = sd.getElement("fieldData")
                        try:
                            field = fd.getElement(0)
                            field_value=field.getValue(0)
                            fieldValues.append(field_value)
                        except:
                            fieldValues.append(None)

            if ev.eventType() == _blpapi.Event.RESPONSE:
                # Response completly received, so we could exit
                break
    finally:
        # Stop the session
        session.stop()
        # return values
        output= _pd.DataFrame.from_records(data=[fieldValues], columns=tickers, index=[targetFieldName])
        if as_columns==False:
            output = output.transpose()

        return output

def BDP_Multiple_Fields(listSecID,listFields,overrideList):
    '''
    Generic reference data request (M securities, N fields). The code works even if there are "bad securities" :keyword "bad fields" (will output as None for those)
    :param listSecID: list of the tickers we want to use, eg ["IBM US Equity", "AAPL US Equity"]
    :param listFields: list of the tickers (eg ["NAME","CUST_TRR_RETURN_HOLDING_PER"])
    :param overrideList: [[overrideParam,overrideValue]] eg [['CUST_TRR_START_DT',startDateYYYYMMDD],['CUST_TRR_END_DT',endDateYYYYMMDD]]
    :return: returns a dataframe with {securities as rows} and {Fields as columns}
    '''
    options = _misc.parseCmdLine()

    # Fill SessionOptions
    sessionOptions = _blpapi.SessionOptions()
    sessionOptions.setServerHost(options.host)
    sessionOptions.setServerPort(options.port)

    print("Connecting to %s:%s" % (options.host, options.port))
    # Create a Session
    session = _blpapi.Session(sessionOptions)

    #initialise some counters for stats
    numFieldsEmptyValues = [0] * len(listFields)
    emptyTickers=[]

    # Start a Session
    if not session.start():
        print("Failed to start session.")
        return

    try:
        # Open service to get historical data from
        if not session.openService("//blp/refdata"):
            print("Failed to open //blp/refdata")
            return

        # Obtain previously opened service
        refDataService = session.getService("//blp/refdata")

        # Create the request for the historical data
        request = refDataService.createRequest("ReferenceDataRequest")

        # Fill the request
        #   a) securities
        for sec_id in listSecID:
            request.getElement("securities").appendValue(sec_id)
        #   b) fields
        for field_id in listFields:
            request.getElement("fields").appendValue(field_id)
        #   d) overrides
        overrides = request.getElement("overrides")
        for ovr in overrideList:
            override = overrides.appendElement()
            override.setElement('fieldId', ovr[0])
            override.setElement('value', ovr[1])

        # Send the request
        if False:
            print("Sending Request:", request)
        else:
            if len(overrideList)>0:
                print("Processing BDP: "+str(len(listSecID))+' securities, '+str(len(listFields))+' fields and '+str(len(overrideList))+' overrides')
        session.sendRequest(request)

        # Process received events
        RESPONSE= _blpapi.Event.RESPONSE
        PARTIAL_RESPONSE= _blpapi.Event.PARTIAL_RESPONSE
        tickers=[]

        securityData=[]
        position = dict(zip(listFields, range(len(listFields))))
        emptyOneSecurity=[None]*len(listFields)

        while(True):
            # We provide timeout to give the chance for Ctrl+C handling:
            ev = session.nextEvent(10)         #original value=500
            for msg in ev:
                #print(msg)
                if (ev.eventType()==PARTIAL_RESPONSE) or (ev.eventType()==RESPONSE):
                    dt = msg.getElement("securityData")
                    numSecurities = dt.numValues()
                    for i in range(0, numSecurities):
                        sd = dt.getValue(i)
                        sn = sd.getElement("security").getValue()
                        tickers.append(sn)
                        fd = sd.getElement("fieldData")
                        fieldValuesOneSecurity = list(emptyOneSecurity)
                        try:
                            numFields=fd.numElements()
                            for j in range(0,numFields):
                                field = fd.getElement(j)
                                field_value=field.getValue(0)
                                field_name=str(field.name())
                                fieldValuesOneSecurity[position[field_name]]=field_value
                            securityData.append(fieldValuesOneSecurity)
                            #record securities with no info
                            if fieldValuesOneSecurity==emptyOneSecurity:
                                emptyTickers.append(sn)
                            else:
                                # record number of empty fields for a valid security
                                for i in range(0, numFields):
                                    if fieldValuesOneSecurity[i] == None:
                                        numFieldsEmptyValues[i] += 1
                        except:
                            securityData.append(emptyOneSecurity)
                            emptyTickers.append(sn)

            if ev.eventType() == _blpapi.Event.RESPONSE:
                # Response completly received, so we could exit
                break
    finally:
        # Stop the session
        session.stop()
        # return values
        output = _pd.DataFrame.from_records(data=securityData, columns=listFields)
        output['security'] = tickers
        output.set_index('security', inplace=True)

        # print stats
        X_Header=True
        if len(emptyTickers)>0:
            print('XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX')
            print('WARNING: '+str(len(emptyTickers))+' BAD securities:')
            print(emptyTickers)
            print('XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX')
            X_Header=False

        for i in range(0,len(listFields)):
            if numFieldsEmptyValues[i]>0:
                if X_Header:
                    print('XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX')
                    X_Header=False
                print('WARNING: Num empty '+listFields[i]+' in good securities = '+str(numFieldsEmptyValues[i]))
                print('XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX')
        return output



def BDP_Dividends(listSecID,overrideList):
    '''
    :param listSecID: list of the tickers we want to use, eg ["IBM US Equity", "AAPL US Equity"]
    :param overrideList: [[overrideParam,overrideValue]] eg [['DVD_START_DT',startDateYYYYMMDD],['DVD_END_DT',endDateYYYYMMDD]]  IF EMPTY -> GETS ALL DIVS
    :return: returns a dataframe with dividends paid by each ticker, each row includes the 7 fields pertaining to dividends
    '''
    dividendFields = ['Declared Date', 'Ex-Date', 'Record Date', 'Payable Date', 'Dividend Amount',
                      'Dividend Frequency', 'Dividend Type']
    targetFieldName='DVD_HIST_ALL'
    dataLineTemplate = _pd.DataFrame(index=[None], columns= (['ticker'] + dividendFields), data=None)
    allDividends = dataLineTemplate.copy()
    options = _misc.parseCmdLine()

    # Fill SessionOptions
    sessionOptions = _blpapi.SessionOptions()
    sessionOptions.setServerHost(options.host)
    sessionOptions.setServerPort(options.port)

    print("Connecting to %s:%s" % (options.host, options.port))
    # Create a Session
    session = _blpapi.Session(sessionOptions)

    # Start a Session
    if not session.start():
        print("Failed to start session.")
        return

    try:
        # Open service to get historical data from
        if not session.openService("//blp/refdata"):
            print("Failed to open //blp/refdata")
            return

        # Obtain previously opened service
        refDataService = session.getService("//blp/refdata")

        # Create the request for the historical data
        request = refDataService.createRequest("ReferenceDataRequest")

        # Fill the request
        #   a) securities
        for sec_id in listSecID:
            request.getElement("securities").appendValue(sec_id)
        #   b) fields
        request.getElement("fields").appendValue(targetFieldName)
        #   d) overrides
        overrides = request.getElement("overrides")
        for ovr in overrideList:
            override = overrides.appendElement()
            override.setElement('fieldId', ovr[0])
            override.setElement('value', ovr[1])

        # Send the request
        if False:
            print("Sending Request:", request)
        else:
            if len(overrideList)>0:
                print("Processing BDP - "+targetFieldName+' - '+str(overrideList[0]))
        session.sendRequest(request)

        # Process received events
        RESPONSE= _blpapi.Event.RESPONSE
        PARTIAL_RESPONSE= _blpapi.Event.PARTIAL_RESPONSE

        while(True):
            # We provide timeout to give the chance for Ctrl+C handling:
            ev = session.nextEvent(10)         #original value=500
            for msg in ev:
                #print(msg)
                if (ev.eventType()==PARTIAL_RESPONSE) or (ev.eventType()==RESPONSE):
                    dt = msg.getElement("securityData")
                    numSecurities = dt.numValues()
                    for i in range(0, numSecurities):
                        sd = dt.getValue(i)
                        sn = sd.getElement("security").getValue()
                        fd = sd.getElement("fieldData")
                        try:
                            field = fd.getElement(0)
                            numDividendInstances = field.numValues()
                            for k in range(0,numDividendInstances):
                                dataLine=dataLineTemplate.copy()
                                dataLine['ticker']=sn
                                field_value=field.getValue(k)
                                for dividendField in dividendFields:
                                    try:
                                        dataLine[dividendField] = field_value.getElement(dividendField).getValue(0)
                                    except:
                                        dataLine[dividendField] = None
                                allDividends = allDividends.append(dataLine,ignore_index=True)
                        except:
                            print('Exception happened')
                            pass

            if ev.eventType() == _blpapi.Event.RESPONSE:
                # Response completly received, so we could exit
                break
    finally:
        # Stop the session
        session.stop()
        # clean the first row and return values
        allDividends=allDividends.dropna(how='all')
        allDividends['Dividend Amount'] = allDividends['Dividend Amount'].round(decimals=10)
        return allDividends


def BDP_Bond_Cashflows(listSecID,overrideList,face_amount=1):
    '''
    :param listSecID: list of the tickers we want to use, eg ["US912810SM18 Govt"] (note that the format " TII 0 1/4 02/15/50 Govt" doesn't work
    :param overrideList: [['SETTLE_DT', YYYYMMDD]] - If not specified it will be from today
    :param face_amount: if unspecified, it will return the cashflows per unit (eg 1 = 100%), otherwise specify eg 1,000,000
    :return: returns a dataframe with coupon and principal amounts paid from the settle date per unit ($1 of face value) or as otherwise specified by each ticker
    '''
    dividendFields = ['Payment Date', 'Coupon Amount', 'Principal Amount']
    targetFieldName='DES_CASH_FLOW'
    cfMultiplier = 1/1000000 * face_amount
    dataLineTemplate = _pd.DataFrame(index=[None], columns= (['ticker'] + dividendFields), data=None)
    allDividends = dataLineTemplate.copy()
    options = _misc.parseCmdLine()

    # Fill SessionOptions
    sessionOptions = _blpapi.SessionOptions()
    sessionOptions.setServerHost(options.host)
    sessionOptions.setServerPort(options.port)

    print("Connecting to %s:%s" % (options.host, options.port))
    # Create a Session
    session = _blpapi.Session(sessionOptions)

    # Start a Session
    if not session.start():
        print("Failed to start session.")
        return

    try:
        # Open service to get historical data from
        if not session.openService("//blp/refdata"):
            print("Failed to open //blp/refdata")
            return

        # Obtain previously opened service
        refDataService = session.getService("//blp/refdata")

        # Create the request for the historical data
        request = refDataService.createRequest("ReferenceDataRequest")

        # Fill the request
        #   a) securities
        for sec_id in listSecID:
            request.getElement("securities").appendValue(sec_id)
        #   b) fields
        request.getElement("fields").appendValue(targetFieldName)
        #   d) overrides
        overrides = request.getElement("overrides")
        for ovr in overrideList:
            override = overrides.appendElement()
            override.setElement('fieldId', ovr[0])
            override.setElement('value', ovr[1])

        # Send the request
        if False:
            print("Sending Request:", request)
        else:
            if len(overrideList)>0:
                print("Processing BDP - "+targetFieldName+' - '+str(overrideList[0]))
        session.sendRequest(request)

        # Process received events
        RESPONSE= _blpapi.Event.RESPONSE
        PARTIAL_RESPONSE= _blpapi.Event.PARTIAL_RESPONSE

        while(True):
            # We provide timeout to give the chance for Ctrl+C handling:
            ev = session.nextEvent(10)         #original value=500
            for msg in ev:
                #print(msg)
                if (ev.eventType()==PARTIAL_RESPONSE) or (ev.eventType()==RESPONSE):
                    dt = msg.getElement("securityData")
                    numSecurities = dt.numValues()
                    for i in range(0, numSecurities):
                        sd = dt.getValue(i)
                        sn = sd.getElement("security").getValue()
                        fd = sd.getElement("fieldData")
                        try:
                            field = fd.getElement(0)
                            numDividendInstances = field.numValues()
                            for k in range(0,numDividendInstances):
                                dataLine=dataLineTemplate.copy()
                                dataLine['ticker']=sn
                                field_value=field.getValue(k)
                                for dividendField in dividendFields:
                                    try:
                                        dataLine[dividendField] = field_value.getElement(dividendField).getValue(0)
                                    except:
                                        dataLine[dividendField] = None
                                dataLine['Coupon Amount'] = dataLine['Coupon Amount'] * cfMultiplier
                                dataLine['Principal Amount'] = dataLine['Principal Amount'] * cfMultiplier
                                allDividends = allDividends.append(dataLine,ignore_index=True)
                        except:
                            print('Exception happened for '+sn)
                            pass

            if ev.eventType() == _blpapi.Event.RESPONSE:
                # Response completly received, so we could exit
                break
    finally:
        # Stop the session
        session.stop()
        # clean the first row and return values
        allDividends=allDividends.dropna(how='all')
        allDividends['Coupon Amount'] = allDividends['Coupon Amount'].round(decimals=10)
        allDividends['Principal Amount'] = allDividends['Principal Amount'].round(decimals=10)
        return allDividends




# SECONDARY FUNCTIONS FOR UNIVERSE GENERATION - NOT USED IN SISTEMA AS FAR AS I CAN SEE!!

def generate_universe(date_list=["20220101","20220630","20221231"], private_screen_name="FTSE100", only_index_names = False):
    '''
    :param date_list: list of dates as string in YYYYMMDD format eg ["20220101","20220630","20221231"]
    :param private_screen_name: name of private Bloomberg screen
    :param only_index_names: If True, then it takes only active index names on that period (eg keep suvivorship bias)
    :return: IndexConstituents -> a df where the columns are the constituents of the screen, the rows each of the dates, and the values whether the security is part of the screen as of that date (1) or not (None)
             weights -> df with unit weights for each date and ticker
             market cap -> df with market cap for each date and ticker
    '''

    date_format = "%Y%m%d"
    universeComposition = []
    allTickers = []

    for refDate in date_list:
        tickerList = BEQS(private_screen_name, "PRIVATE", refDate)
        allTickers = allTickers + list(set(tickerList) - set(allTickers))  # adds unique elements to allTickers
        universeComposition.append((refDate, tickerList))

    #zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz
    # INDEX CONSTITUENTS

    indexConstituents = _pd.DataFrame(columns=allTickers)
    for (endDateYYYYMMDD,listSecID) in universeComposition:
        row_1 = _pd.DataFrame(columns=listSecID, data=_np.ones(shape=(1, len(listSecID))))
        indexConstituents = _pd.concat([indexConstituents, row_1], axis=0, ignore_index=True)
    indexConstituents['dates']=date_list
    indexConstituents.set_index('dates',inplace=True)

    # return indexConstituents
    # zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz

    #zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz
    # INDEX WEIGHTS

    df = _pd.DataFrame(columns=indexConstituents.columns)
    df_raw=df.copy()

    for (endDateYYYYMMDD,row) in indexConstituents.iterrows():
        endDateYYYYMMDD=str(endDateYYYYMMDD)
        # cambio 10-jun 12:56
        if only_index_names:
            listSecID=list(row[row==1].index) #esto selecciona solo los tickers que, en ese mes, forman parte del indice
        else:
            listSecID = list(row.index) #coge todos los tickers con independencia de si estan en el indice o no
        # -------------------

        listFieldID = ["CUR_MKT_CAP"]
        listFieldScale = ["RAW"]
        periodicityAdjustment = "ACTUAL"
        periodicitySelection = "DAILY"
        startDateYYYYMMDD = (_datetime.datetime.strptime(endDateYYYYMMDD, date_format) - _datetime.timedelta(days=3)).strftime(date_format)
        maxDataPoints = 1
        convertDatesToNumbers = False
        oneTablePer = "FIELD"  # "SECURITY" or "FIELD"

        response =BDH(listSecID=listSecID,
                      listFieldID=listFieldID,
                      listFieldScale=listFieldScale,
                      startDateYYYYMMDD=startDateYYYYMMDD,
                      endDateYYYYMMDD=endDateYYYYMMDD,
                      periodicityAdjustment=periodicityAdjustment,
                      periodicitySelection=periodicitySelection,
                      maxDataPoints=maxDataPoints,
                      listOverrides=[],
                      convertDatesToNumbers=convertDatesToNumbers,
                      oneTablePer=oneTablePer,
                      onlyCommonDates=False,
                      return_as_dict=False)

        totalCap = response[0].sum(1)[0]

        raw_cap=response[0].copy()
        raw_cap['date'] = response[0].index

        response[0] = response[0] / totalCap
        response[0]['date'] = response[0].index

        df = _pd.concat([df, response[0]], axis=0, ignore_index=True)
        df_raw= _pd.concat([df_raw, raw_cap], axis=0, ignore_index=True)
        print('Market cap imported for ' + response[0]['date'][0].strftime(date_format))

    weights = df.set_index('date').fillna(0)
    market_caps=df_raw.set_index('date').fillna(0)

    return [indexConstituents,weights,market_caps]


def BDS(security="CLA COMB Comdty",targetFieldName="FUT_CHAIN",overrideList=[["CHAIN_DATE","20220127"]]):
    '''
    :param security: eg "CLA COMB Comdty"
    :param targetFieldName: eg "FUT_CHAIN" , "OPT_CHAIN"
    :param overrideList: eg [["CHAIN_DATE","20220127"]] - the asof date in YYYYMMDD
    :return: returns a dataframe with the given futures or option chain for that particular date, sorted by order of contract (1st first)
    '''
    options = _misc.parseCmdLine()

    # Fill SessionOptions
    sessionOptions = _blpapi.SessionOptions()
    sessionOptions.setServerHost(options.host)
    sessionOptions.setServerPort(options.port)

    print("Connecting to %s:%s" % (options.host, options.port))
    # Create a Session
    session = _blpapi.Session(sessionOptions)

    # Start a Session
    if not session.start():
        print("Failed to start session.")
        return

    try:
        # Open service to get historical data from
        if not session.openService("//blp/refdata"):
            print("Failed to open //blp/refdata")
            return

        # Obtain previously opened service
        refDataService = session.getService("//blp/refdata")

        # Create the request for the historical data
        request = refDataService.createRequest("ReferenceDataRequest")

        # Fill the request
        #   a) securities
        request.getElement("securities").appendValue(security)
        #   b) fields
        request.getElement("fields").appendValue(targetFieldName)
        #   d) overrides
        overrides = request.getElement("overrides")
        for ovr in overrideList:
            override = overrides.appendElement()
            override.setElement('fieldId', ovr[0])
            override.setElement('value', ovr[1])

        # Send the request
        if False:
            print("Sending Request:", request)
        else:
            if len(overrideList)>0:
                print("Processing BDP - "+targetFieldName+' - '+str(overrideList[0]))
        session.sendRequest(request)

        # Process received events
        RESPONSE= _blpapi.Event.RESPONSE
        PARTIAL_RESPONSE= _blpapi.Event.PARTIAL_RESPONSE
        tickers=[]
        fieldValues=[]

        while(True):
            # We provide timeout to give the chance for Ctrl+C handling:
            ev = session.nextEvent(10)         #original value=500
            for msg in ev:
                #print(msg)
                if (ev.eventType()==PARTIAL_RESPONSE) or (ev.eventType()==RESPONSE):
                    dt = msg.getElement("securityData")
                    numSecurities = dt.numValues()
                    for i in range(0, numSecurities):
                        sd = dt.getValue(i)
                        sn = sd.getElement("security").getValue()
                        tickers.append(sn)
                        fd = sd.getElement("fieldData")
                        try:
                            field = fd.getElement(0)
                            # field_value=field.getValue(0)
                            for f in field:
                                fieldValues.append(f.getElementValue("Security Description"))
                        except:
                            fieldValues.append(None)

            if ev.eventType() == _blpapi.Event.RESPONSE:
                # Response completly received, so we could exit
                break
    finally:
        # Stop the session
        session.stop()
        # return values
        output = _pd.DataFrame.from_records({tickers[0]:fieldValues})

        return output